#include "Hypervisor.h"

#include <ntifs.h>

#include "MSR.h"
#include "CPUID.h"

#include <intrin.h>

namespace
{
    enum class CPU_VENDOR {
        cpuIntel,
        cpuAmd,
        cpuUnknown
    };

    CPU_VENDOR GetCpuVendor()
    {
        static CPU_VENDOR CpuVendor = CPU_VENDOR::cpuUnknown;
        if (CpuVendor != CPU_VENDOR::cpuUnknown)
        {
            return CpuVendor;
        }
        CPUID_REGS Regs;
        __cpuid(Regs.Raw, CPUID::Generic::CPUID_MAXIMUM_FUNCTION_NUMBER_AND_VENDOR_ID);
        if (Regs.Regs.Ebx == 'uneG' && Regs.Regs.Edx == 'Ieni' && Regs.Regs.Ecx == 'letn')
        {
            CpuVendor = CPU_VENDOR::cpuIntel;
        }
        else if (Regs.Regs.Ebx == 'htuA' && Regs.Regs.Edx == 'itne' && Regs.Regs.Ecx == 'DMAc')
        {
            CpuVendor = CPU_VENDOR::cpuAmd;
        }

        return CpuVendor;
    }
}

namespace VMX
{
    using namespace Intel;
    
    static bool VirtualizeAllProcessors()
    {
        /* We're good to go */
        return false;
    }

    static bool IsVmxSupported()
    {
        CPUID_REGS Regs = {};

        // Check the 'GenuineIntel' vendor name:
        __cpuid(Regs.Raw, CPUID::Generic::CPUID_MAXIMUM_FUNCTION_NUMBER_AND_VENDOR_ID);

        if (Regs.Regs.Ebx != 'uneG' || Regs.Regs.Edx != 'Ieni' || Regs.Regs.Ecx != 'letn')
        {
            return false;
        }

        // Support by processor:
        __cpuid(Regs.Raw, CPUID::Intel::CPUID_FEATURE_INFORMATION);
        if (!reinterpret_cast<CPUID::FEATURE_INFORMATION*>(&Regs)->Intel.VMX)
        {
            return false;
        }

        // Check the VMX is locked in BIOS:
        IA32_FEATURE_CONTROL MsrFeatureControl = {};
        MsrFeatureControl.Value = __readmsr(static_cast<unsigned long>(INTEL_MSR::IA32_FEATURE_CONTROL));
        
        return MsrFeatureControl.Bitmap.LockBit == TRUE;
    }
}

namespace Hypervisor
{
    bool Virtualize()
    {
        CPU_VENDOR CpuVendor = GetCpuVendor();
        if (CpuVendor == CPU_VENDOR::cpuUnknown) return false;
        
        bool Status = false;
        
        switch (CpuVendor) {
        case CPU_VENDOR::cpuIntel:
        {
            if (!VMX::IsVmxSupported()) return false;
            Status = VMX::VirtualizeAllProcessors();
            break;
        }
        case CPU_VENDOR::cpuAmd:
        {
            /* Not supported yet */
            break;
        }
        }

        return Status;
    }

    bool Devirtualize()
    {
        return false;
    }
}